#include "pch.hpp"
#include "SwrAddrDef.h"
#include "chalcopyrite.hpp"
#include "ProfileIO.hpp"

#define MINIMAL_USE_PROCESSHEAPSTRING
#include "MinimalPath.hpp"

#define POLL_INTERVAL 100


#define DEF_SWR_WINDOW_CLASS "th105_106"
#define DEF_SWR_WINDOW_TEXT  "zV Ver1.06"

#define DEF_SWR_ADDR_PBATTLEMGR 0x006E6244
#define DEF_SWR_ADDR_PNETOBJECT 0x006E62FC
#define DEF_SWR_ADDR_COMMMODE   0x006E62EC
#define DEF_SWR_ADDR_LCHARID    0x006E6FF0
#define DEF_SWR_ADDR_RCHARID    0x006E7010
#define DEF_SWR_ADDR_SCENEID    0x006ECE7C

#define DEF_SWR_ADDR_LCHAROFS	0x0c
#define DEF_SWR_ADDR_RCHAROFS	0x10
#define DEF_SWR_ADDR_BTLMODEOFS	0x88

#define DEF_SWR_ADDR_LPROFOFS   0x04
#define DEF_SWR_ADDR_RPROFOFS   0x24

#define DEF_SWR_ADDR_NETUDPOFS  0x3B4

/* For Server (it may be also used in client mode.) */
#define DEF_SWR_ADDR_ADRBEGOFS  (DEF_SWR_ADDR_NETUDPOFS + 0xFC)
/* For Client */
#define DEF_SWR_ADDR_TOADDROFS  (DEF_SWR_ADDR_NETUDPOFS + 0x2C)

#define DEF_SWR_ADDR_WINCNTOFS  0x4E8

static char SWR_WINDOW_CLASS[256];
static char SWR_WINDOW_TEXT[256];

static DWORD SWR_ADDR_PBATTLEMGR;
static DWORD SWR_ADDR_PNETOBJECT;
static DWORD SWR_ADDR_COMMMODE;
static DWORD SWR_ADDR_LCHARID;
static DWORD SWR_ADDR_RCHARID;
static DWORD SWR_ADDR_SCENEID;

static DWORD SWR_ADDR_LCHAROFS;
static DWORD SWR_ADDR_RCHAROFS;
static DWORD SWR_ADDR_BTLMODEOFS;

static DWORD SWR_ADDR_LPROFOFS;
static DWORD SWR_ADDR_RPROFOFS;

static DWORD SWR_ADDR_NETUDPOFS;

/* For Server (it may be also used in client mode.) */
static DWORD SWR_ADDR_ADRBEGOFS;
/* For Client */
static DWORD SWR_ADDR_TOADDROFS;

static DWORD SWR_ADDR_WINCNTOFS;

static enum SFMSTATE {
	SFM_FINDWINDOW,
	SFM_WATCHSWR,
};

static HANDLE s_ThProc;
static HWND   s_ThWnd;

static HWND s_callbackWnd;
static int  s_callbackMsg;

static HANDLE s_userThread;
static HANDLE s_thread;
static DWORD s_threadId;

static int s_paramOld[SWRPARAM_MAX];
static char s_paramStr[260];

static void SwrLoadProfile()
{
#define LoadAddress(N) N = profile.ReadInteger("SwrAddress", #N, DEF_##N)
#define LoadChar(N) profile.ReadString("SwrName", #N, DEF_##N, N, sizeof N)

	Minimal::ProcessHeapPathA profPath = g_appPath;
	CProfileIO profile(profPath /= "SwrAddr.ini");
	LoadChar(SWR_WINDOW_CLASS);
	LoadChar(SWR_WINDOW_TEXT);

	LoadAddress(SWR_ADDR_PBATTLEMGR);
	LoadAddress(SWR_ADDR_PNETOBJECT);
	LoadAddress(SWR_ADDR_COMMMODE);
	LoadAddress(SWR_ADDR_LCHARID);
	LoadAddress(SWR_ADDR_RCHARID);
	LoadAddress(SWR_ADDR_SCENEID);

	LoadAddress(SWR_ADDR_LCHAROFS);
	LoadAddress(SWR_ADDR_RCHAROFS);
	LoadAddress(SWR_ADDR_BTLMODEOFS);

	LoadAddress(SWR_ADDR_LPROFOFS);
	LoadAddress(SWR_ADDR_RPROFOFS);

	LoadAddress(SWR_ADDR_NETUDPOFS);

	LoadAddress(SWR_ADDR_ADRBEGOFS);
	LoadAddress(SWR_ADDR_TOADDROFS);

	LoadAddress(SWR_ADDR_WINCNTOFS);
}

static void APIENTRY SwrAPCCallback(ULONG_PTR)
{
}

static void SwrCallback(short Msg, short param1, int param2)
{
	if(s_callbackWnd) {
		::PostMessage(s_callbackWnd, s_callbackMsg, MAKELONG(Msg, param1), param2);
	}
}


static SFMSTATE SwrStateFindWindow()
{
	s_ThWnd = ::FindWindow(SWR_WINDOW_CLASS, SWR_WINDOW_TEXT);
	if(s_ThWnd != NULL) {
		DWORD dwThId;
		::GetWindowThreadProcessId(s_ThWnd, &dwThId);
		s_ThProc = ::OpenProcess(PROCESS_ALL_ACCESS, FALSE, dwThId);
		SwrCallback(SWRMSG_STATECHANGE, SWRSTATE_WATCH, 0);
		::ZeroMemory(s_paramOld, sizeof s_paramOld);
		return SFM_WATCHSWR;
	} else {
		return SFM_FINDWINDOW;
	}
}

static SFMSTATE SwrStateWatchSwr()
{
	DWORD ret = ::WaitForSingleObject(s_ThProc, 0);
	if(ret != WAIT_TIMEOUT) {
		::CloseHandle(s_ThProc);
		s_ThProc = NULL;
		SwrCallback(SWRMSG_STATECHANGE, SWRSTATE_NOTFOUND, 0);
		return SFM_FINDWINDOW;
	}

	int param;
	for(int i = 0; i < SWRPARAM_MAX; ++i) {
		if((param = SwrGetParam(i)) != -1) {
			if(param != s_paramOld[i])
				SwrCallback(SWRMSG_PARAMCHANGE, i, param);
			s_paramOld[i] = param;
		} else s_paramOld[i] = 0;
	}
	return SFM_WATCHSWR;
}

static DWORD WINAPI SwrAddrWorkThread(LPVOID)
{
	MSG msg;
	SFMSTATE sfmState = SFM_FINDWINDOW;

	while(!::PeekMessage(&msg, NULL, 0, 0, PM_REMOVE) || msg.message != WM_QUIT) {
		switch(sfmState) {
		case SFM_FINDWINDOW: sfmState = SwrStateFindWindow(); break;
		case SFM_WATCHSWR:   sfmState = SwrStateWatchSwr();   break;
		};
		::SleepEx(POLL_INTERVAL, TRUE);
	}
	::ExitThread(0);
	return 0;
}

int SwrAddrInit(HWND callbackWnd, int callbackMsg)
{
	SwrLoadProfile();
	s_callbackWnd = callbackWnd;
	s_callbackMsg = callbackMsg;
	s_thread = ::CreateThread(NULL, 0, SwrAddrWorkThread, NULL, 0, &s_threadId);
	s_userThread = ::GetCurrentThread();
	return s_thread != NULL;
}

int SwrAddrFinish()
{
	if(s_thread) {
		::PostThreadMessage(s_threadId, WM_QUIT, 0, 0);
		::QueueUserAPC(SwrAPCCallback, s_thread, 0);
		::WaitForSingleObject(s_thread, INFINITE);
		::CloseHandle(s_thread);
	}
	return 0;
}

int SwrGetParam(int param)
{
	DWORD dwReadSize;
	DWORD dwordData, dwordData2;
	BYTE byteData;

	switch(param) {
	case SWRPARAM_SCENE:
		if(::ReadProcessMemory(s_ThProc, (LPVOID)SWR_ADDR_SCENEID, &dwordData, 4, &dwReadSize)) {
			return dwordData;
		}
		break;
	case SWRPARAM_COMMMODE:
		if(::ReadProcessMemory(s_ThProc, (LPVOID)SWR_ADDR_COMMMODE, &dwordData, 4, &dwReadSize)) {
			return dwordData;
		}
		break;
	case SWRPARAM_BATTLEMODE:
		if(::ReadProcessMemory(s_ThProc, (LPVOID)SWR_ADDR_PBATTLEMGR, &dwordData, 4, &dwReadSize) &&
		   ::ReadProcessMemory(s_ThProc, (LPVOID)(dwordData + SWR_ADDR_BTLMODEOFS), &dwordData, 4, &dwReadSize)) {
			return dwordData;
		}
		break;
	case SWRPARAM_LCHARID:
		if(::ReadProcessMemory(s_ThProc, (LPVOID)SWR_ADDR_LCHARID, &dwordData, 4, &dwReadSize)) {
			return dwordData;
		}
		break;
	case SWRPARAM_RCHARID:
		if(::ReadProcessMemory(s_ThProc, (LPVOID)SWR_ADDR_RCHARID, &dwordData, 4, &dwReadSize)) {
			return dwordData;
		}
		break;
	case SWRPARAM_LWINCOUNT:
		if(::ReadProcessMemory(s_ThProc, (LPVOID)SWR_ADDR_PBATTLEMGR, &dwordData, 4, &dwReadSize) &&
		   ::ReadProcessMemory(s_ThProc, (LPVOID)(dwordData + SWR_ADDR_LCHAROFS), &dwordData, 4, &dwReadSize) &&
		   ::ReadProcessMemory(s_ThProc, (LPVOID)(dwordData + SWR_ADDR_WINCNTOFS), &byteData, 1, &dwReadSize)) {
			return byteData;
		}
		break;
	case SWRPARAM_RWINCOUNT:
		if(::ReadProcessMemory(s_ThProc, (LPVOID)SWR_ADDR_PBATTLEMGR, &dwordData, 4, &dwReadSize) &&
		   ::ReadProcessMemory(s_ThProc, (LPVOID)(dwordData + SWR_ADDR_RCHAROFS), &dwordData, 4, &dwReadSize) &&
		   ::ReadProcessMemory(s_ThProc, (LPVOID)(dwordData + SWR_ADDR_WINCNTOFS), &byteData, 1, &dwReadSize)) {
			return byteData;
		}
		break;
	case SWRPARAM_LPROFNAME:
		if(::ReadProcessMemory(s_ThProc, (LPVOID)SWR_ADDR_PNETOBJECT, &dwordData, 4, &dwReadSize) &&
			::ReadProcessMemory(s_ThProc, (LPVOID)(dwordData + SWR_ADDR_LPROFOFS), s_paramStr, SWR_ADDR_PROFSIZE, &dwReadSize)) {
			return (int)s_paramStr;
		}
		break;
	case SWRPARAM_RPROFNAME:
		if(::ReadProcessMemory(s_ThProc, (LPVOID)SWR_ADDR_PNETOBJECT, &dwordData, 4, &dwReadSize) &&
			::ReadProcessMemory(s_ThProc, (LPVOID)(dwordData + SWR_ADDR_RPROFOFS), s_paramStr, SWR_ADDR_PROFSIZE, &dwReadSize)) {
			return (int)s_paramStr;
		}
		break;
	case SWRPARAM_TOADDR_CLIENT:
		if(::ReadProcessMemory(s_ThProc, (LPVOID)SWR_ADDR_PNETOBJECT, &dwordData, 4, &dwReadSize) &&
		   ::ReadProcessMemory(s_ThProc, (LPVOID)(dwordData + SWR_ADDR_TOADDROFS), &dwordData2, 4, &dwReadSize) && dwordData2 != 0 &&
		   ::ReadProcessMemory(s_ThProc, (LPVOID)(dwordData + SWR_ADDR_TOADDROFS + 4), &dwordData2, 4, &dwReadSize)) {
			return dwordData2;
		}
	case SWRPARAM_TOADDR_SERVER:
		if(::ReadProcessMemory(s_ThProc, (LPVOID)SWR_ADDR_PNETOBJECT, &dwordData, 4, &dwReadSize) &&
		   ::ReadProcessMemory(s_ThProc, (LPVOID)(dwordData + SWR_ADDR_ADRBEGOFS), &dwordData, 4, &dwReadSize) &&
		   ::ReadProcessMemory(s_ThProc, (LPVOID)(dwordData + 0x10), &dwordData2, 4, &dwReadSize) && dwordData2 != -1 &&
		   ::ReadProcessMemory(s_ThProc, (LPVOID)(dwordData + 0x04), &dwordData, 4, &dwReadSize)) {
			return dwordData;
		}
		break;
	}
	return -1;
}
